from discminer.mining_control import _mining_radial_profiles
from discminer.mining_utils import (init_data_and_model,
                                    get_noise_mask,
                                    get_1d_plot_decorators,
                                    load_moments,
                                    load_disc_grid,
                                    show_output,
                                    MEDIUM_SIZE)

from discminer.rail import Rail
from discminer.plottools import (make_up_ax,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 make_1d_legend,                                 
                                 make_substructures,
                                 use_discminer_style)

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from scipy.signal import savgol_filter
from scipy.interpolate import interp1d

from astropy import units as u
import json

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_radial_profiles(None)
    args = parser.parse_args()

figsize1row = (11,4)
figsize2row = (11,6)    

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Mstar = best['velocity']['Mstar']
vsys = best['velocity']['vsys']
vel_sign = best['velocity']['vel_sign']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']

gaps = custom['gaps']
rings = custom['rings']
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

clabel, clabel_res, clim0, clim0_res, clim1, clim1_res, unit = get_1d_plot_decorators(args.moment)

#*******************
#LOAD DISC GEOMETRY
#*******************
R, phi, z = load_disc_grid()

R_s = R[args.surface]*u.m.to('au')
phi_s = phi[args.surface]

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2,
                                  mask_phi={'map2d': np.degrees(phi_s),
                                            'lims': args.mask_phi},
                                  mask_R={'map2d': R_s,
                                          'lims': args.mask_R})
vchannels = datacube.vchannels
model.make_model()

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)
ref_surf = mtags['ref_surf']
tag_base = mtags['base']

#**************************
#ABSOLUTE RESIDUALS    
if args.moment=='velocity':
    residuals_abs = np.abs(moment_data-vsys) - np.abs(moment_model-vsys)

#**************************
#RADIAL BINS AND UTILS
beam_au = datacube.beam_size.to('au').value
    
if args.absolute_Rinner>=0:
    Rprof_in = args.absolute_Rinner
else:
    Rprof_in = args.Rinner*beam_au

if args.absolute_Router>=0:
    Rprof_out = args.absolute_Router
else:
    Rprof_out = args.Router*Rout
    
R_prof = np.arange(Rprof_in, Rprof_out, beam_au/4.0) #changed to beam/4 from beam/5 before
xlim0, xlim1 = 0.5*R_prof[0], 1.05*R_prof[-1]

def writetxt(arr, tag=''):
    if tag!='': tag = '_'+tag
    arr = np.asarray(arr).T
    np.savetxt('radial_profile_%s%s.dat'%(args.moment, tag), arr, fmt='%.6f', header='R[au] attribute stddev')

def get_normalisation(mask_ang, component='z'):
    """
    mask_ang : +- mask around disc minor axis
    component : velocity component to get normalisation for    
    """
    if component=='phi':
        div_factor = 4*np.sin(np.pi/2 - np.radians(mask_ang)) * np.abs(np.sin(incl))/(2*np.pi - 4*np.radians(mask_ang))
    elif component=='z':
        div_factor = -np.cos(np.abs(incl))        
    else: return comp
    
    return div_factor

def make_savgol(prof):
    if not args.savgol_filter:
        return prof, None
    try:
        ysav = savgol_filter(prof, 5, 3)
        ysav_deriv = savgol_filter(prof, 5, 3, deriv=1)
    except np.linalg.LinAlgError:
        ysav = prof
        ysav_deriv = None
    return ysav, ysav_deriv 

def make_basic_layout(ax):
    ax.set_xlim(xlim0, xlim1)
    ax.set_xlabel('Radius [au]', fontsize=MEDIUM_SIZE-1)
    mod_major_ticks(ax, axis='x', nbins=10)
    mod_minor_ticks(ax)
    ax.axhline(0, lw=2, ls='--', color='0.7')

def make_profile(ax, R_prof, ysav, y, yerr, kind='data', perr='bar', **kwargs):
    if kind=='data':
        label = 'Data'
        color = 'tomato'
        zorder = 9
        lw = 3.5        
    elif kind=='residuals':
        label = 'Residuals'
        color = '0.1'
        zorder = 9
        lw = 3.5
    elif kind=='model':
        label = 'Model'
        color = 'dodgerblue'
        zorder = 8
        lw = 3.5
    elif kind=='vphi':
        label = r'$\Delta\upsilon_\phi$'
        color = 'dodgerblue'
        zorder = 9
        lw = 3.0
    elif kind=='vz':
        label = r'$\upsilon_{\rm z}$'
        color = 'k'
        zorder = 8
        lw = 3.0
    elif kind=='vr':
        label = r'$\upsilon_{\rm R}$'
        color = '#FFB000'
        zorder = 7
        lw = 3.0
    else:
        raise ValueError(kind)

    ax.plot(R_prof, ysav, c=color, lw=lw, label=label, zorder=zorder)

    if kind!='model':
        if perr=='bar':
            ax.errorbar(R_prof, y, yerr=yerr, c=color,
                        elinewidth=1.2, capsize=1.8, linestyle='none',
                        marker='o', ms=6.5, markeredgewidth=1.7, markerfacecolor='0.8',
                        alpha=0.7, zorder=zorder)
        elif perr=='fill':
            ax.fill_between(R_prof, y+yerr, y-yerr, color=color, alpha=0.15, zorder=zorder)
        else:
            raise ValueError(perr)
        
    if args.writetxt: writetxt([R_prof, y, yerr], tag=kind)

#*************
#MAIN BODY
#*************
kw_ylabel = dict(fontsize=MEDIUM_SIZE-2, labelpad=0)

if args.moment=='velocity':

    mask_ang = args.mask_minor #+-mask around disc minor axis

    #*******************
    #VELOCITY COMPONENTS
    #*******************    
    """    
    if args.surface in ['low', 'lower']:
        mask_r = mask | (np.abs(phi_s) < np.radians(args.mask_major)) | (np.abs(phi_s) > np.radians(180-args.mask_major))
        moment_data = np.where(mask_r, np.nan, moment_data)
        if args.moment=='velocity':
            pass
            moment_data = np.where(np.abs(moment_data-vsys)>5, np.nan, moment_data)    
    """
    if args.kernel in ['doublebell', 'doublegaussian']:
        mask_map_ref = np.abs((moment_data-vsys)/(np.cos(model.projected_coords['phi'][args.surface])*np.sin(incl)))
    else:
        mask_map_ref = None

    kw_avg = dict(surface=ref_surf, av_func=np.nanmean, mask_ang=mask_ang, sigma_thres=args.sigma, mask_from_map=mask_map_ref, interpgrid=args.interpgrid)
    
    #VZ
    rail_vz = Rail(model, residuals, R_prof)
    vel_z, vel_z_error = rail_vz.get_average(tag='vz', **kw_avg)
    div_factor_z = get_normalisation(mask_ang, component='z')

    vel_z /= div_factor_z
    vel_z_error /= div_factor_z 

    #DVPHI
    rail_phi = Rail(model, residuals_abs, R_prof)
    vel_phi, vel_phi_error = rail_phi.get_average(**kw_avg)
    div_factor_phi = get_normalisation(mask_ang, component='phi')

    vel_phi /= div_factor_phi
    vel_phi_error /= div_factor_phi 

    #VPHI
    rail_phi = Rail(model, np.abs(moment_data-vsys), R_prof)
    vel_rot, vel_rot_error = rail_phi.get_average(tag=tag_base+'_vphi_data', plot_diagnostics=True, forward_error=True, **kw_avg)
    vel_rot /= div_factor_phi
    vel_rot_error = vel_phi_error

    #VR
    mask_r = mask | (np.abs(phi_s) < np.radians(args.mask_major)) | (np.abs(phi_s) > np.radians(180-args.mask_major))
    moment_data_r = np.where(mask_r, np.nan, moment_data) 

    f_vp = interp1d(R_prof, vel_rot)
    f_vz = interp1d(R_prof, vel_z)
    R_interp = np.where((R_s>R_prof[0]) & (R_s<R_prof[-1]), R_s, R_prof[0])

    vr = -1/(np.sin(phi_s)*np.sin(incl)) * (moment_data_r - vsys - vel_sign*f_vp(R_interp)*np.cos(phi_s)*np.sin(incl) + f_vz(R_interp)*np.cos(incl))
    vr = np.where(np.abs(vr)>5, np.nan, vr) #hot pixels
    
    rail_vr = Rail(model, vr, R_prof)
    vel_r, vel_r_error = rail_vr.get_average(mask_ang=0.0, surface=ref_surf, av_func=np.nanmedian, interpgrid=args.interpgrid)
    
    #Perfect Keplerian rotation
    coords = {'R': R_prof*u.au.to('m')}    
    velocity_upper = model.get_attribute_map(coords, 'velocity', surface=ref_surf) * vel_sign   

    #PLOT 3D VELOCITIES
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=figsize1row)

    if args.vphi_discminer:
        ysav_phi, ysav_phi_deriv = make_savgol(vel_phi)
        make_profile(ax, R_prof, ysav_phi, vel_phi, vel_phi_error, kind='vphi', perr='fill') #vphi_data-vphi_model   
        ax.scatter(R_prof, vel_phi, ec='dodgerblue', fc='none', s=25, zorder=12)
    else:
        ysav_phi, ysav_phi_deriv = make_savgol(vel_rot-velocity_upper)    
        make_profile(ax, R_prof, ysav_phi, vel_rot-velocity_upper, vel_phi_error, kind='vphi', perr='fill') #vphi_data-vkep
        ax.scatter(R_prof, vel_rot-velocity_upper, ec='dodgerblue', fc='none', s=25, zorder=12)
       
    ysav_z, ysav_z_deriv = make_savgol(vel_z)    
    make_profile(ax, R_prof, ysav_z, vel_z, vel_z_error, kind='vz', perr='fill')
    ax.scatter(R_prof, vel_z, ec='k', fc='none', s=25, zorder=11)
    
    ysav_r, ysav_r_deriv = make_savgol(vel_r)
    make_profile(ax, R_prof, ysav_r, vel_r, vel_r_error, kind='vr', perr='fill')    
    ax.scatter(R_prof, vel_r, ec='#FFB000', fc='none', s=25, zorder=10)
    
    #DECORATIONS
    make_basic_layout(ax)
    ax.set_ylabel(r'$\delta\upsilon$ [km/s]', fontsize=MEDIUM_SIZE, labelpad=10)
    ax.set_ylim(clim0_res, clim1_res)
    make_1d_legend(ax, fontsize=MEDIUM_SIZE+1)    
    make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)
    
    plt.savefig('velocity_components_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    show_output(args)
    
    #*******************
    #ROTATION CURVE
    #*******************
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=figsize1row)
        
    #DATA CURVE
    ysav_rot, ysav_rot_deriv = make_savgol(vel_rot)
    make_profile(ax, R_prof, ysav_rot, vel_rot, vel_rot_error, kind='data')

    #MODEL CURVE
    rail_phi = Rail(model, np.abs(moment_model-vsys), R_prof)
    #vel_phi, _ = rail_phi.get_average(mask_ang=mask_ang, surface=ref_surf)
    vel_phi, _ = rail_phi.get_average(surface=ref_surf, av_func=np.nanmean, mask_ang=mask_ang, sigma_thres=np.inf,
                                      mask_from_map=mask_map_ref, tag=tag_base+'_vphi_model', interpgrid=args.interpgrid, plot_diagnostics=True)
    div_factor_model = get_normalisation(mask_ang, component='phi')
    vel_phi /= div_factor_model
    ysav_phi, ysav_phi_deriv_mod = make_savgol(vel_phi)
    make_profile(ax, R_prof, ysav_phi, vel_phi, _, kind='model')
    
    #PERFECT KEPLERIAN
    coords = {'R': R_prof*u.au.to('m')}
    velocity_upper = model.get_attribute_map(coords, 'velocity', surface=ref_surf) * vel_sign
    ax.plot(R_prof, velocity_upper, c='k', lw=2.5, ls='--', label=r'Keplerian (%.2f Msun)'%Mstar, zorder=13)

    #DECORATIONS
    make_basic_layout(ax)
    ax.set_ylabel(r'Rotation velocity [km/s]', fontsize=MEDIUM_SIZE, labelpad=10)
    ax.set_ylim(clim0, 1.2*np.nanmax(vel_phi))
    make_1d_legend(ax, fontsize=MEDIUM_SIZE-3)
    make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)  
    
    plt.savefig('rotation_curve_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    show_output(args)
    
    """
    #Data rotation curve - perfect_Keplerian
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,4)) 
    ax.plot(R_prof, vel_rot-velocity_upper, c='k', lw=2.5, ls='--', label=r'Keplerian (%.2f Msun)'%Mstar, zorder=13)
    plt.show()
    """
    
else:

    mask_ang = args.mask_minor
    
    if args.surface in ['low', 'lower']:
        mask_r = mask | (np.abs(phi_s) < np.radians(args.mask_major)) | (np.abs(phi_s) > np.radians(180-args.mask_major)) | (moment_data>40)
        moment_data = np.where(mask_r, np.nan, moment_data)

    if args.kernel in ['doublebell', 'doublegaussian']:
        mask_map_ref = moment_data
    else:
        mask_map_ref = None

    #*****************
    #ABSOLUTE PROFILES
    #*****************
    kw_avg = dict(surface=ref_surf, av_func=np.nanmedian, mask_ang=mask_ang, sigma_thres=args.sigma, mask_from_map=mask_map_ref, interpgrid=args.interpgrid, plot_diagnostics=True)
        
    fig, ax = plt.subplots(ncols=1, nrows=2, figsize=figsize2row)

    #DATA CURVE
    rail = Rail(model, moment_data, R_prof)
    radprof, radprof_error = rail.get_average(**kw_avg)

    ysav, ysav_deriv = make_savgol(radprof)    
    make_profile(ax[0], R_prof, ysav, radprof, radprof_error, kind='data')    

    #MODEL CURVE
    rail = Rail(model, moment_model, R_prof)
    radprof, radprof_error = rail.get_average(**kw_avg)

    ysav, ysav_deriv = make_savgol(radprof)
    make_profile(ax[0], R_prof, ysav, radprof, radprof_error, kind='model')        

    #RESIDUALS
    rail = Rail(model, residuals, R_prof)
    radprof, radprof_error = rail.get_average(**kw_avg)

    ysav, ysav_deriv = make_savgol(radprof)
    make_profile(ax[1], R_prof, ysav, radprof, radprof_error, kind='residuals')

    #***********
    #DECORATIONS
    #***********
    ax[0].set_ylim(clim0, clim1)
    ax[1].set_ylim(clim0_res, clim1_res)

    #fmt (Homogenise fmt for both panels)
    #******    
    ticks = ax[1].get_yticks()

    makeint = False
    for tick in ticks:
        if str(tick)[-2:] != '.0':
            break
        else:
            makeint = True

    #ftick = round(ticks[0], 2)
    ftick = round(clim0_res, 2)
    ftick = int(ftick) if makeint else ftick
    ftick_res = str(ftick)
    
    isfloat = isinstance(ftick, float)
    #lfmt = len(str(ftick))
    lfmt = 5
    
    if isfloat:
        if abs(ftick)<0.2 and lfmt<5: #e.g. -0.1
            ftick_res += '0'
            lfmt+=1
        
        ndec = len(str(ftick).split('.')[-1])
        ndec_res = len(str(ftick_res).split('.')[-1])
        
        cfmt = '%'+'%d.%df'%(lfmt+ndec, ndec)
        cfmt_res = '%'+'%d.%df'%(lfmt+ndec_res, ndec_res)        
    else:
        cfmt = cfmt_res = '%'+'%dd'%lfmt
    
    #******
    #Axes
    #****
    for axi in ax:
        make_basic_layout(axi)
        axi.tick_params(which='both', labelsize=MEDIUM_SIZE-3)

    ax[0].yaxis.set_major_formatter(FormatStrFormatter(cfmt))
    ax[1].yaxis.set_major_formatter(FormatStrFormatter(cfmt_res))            
    ax[0].set_xlabel(None)
    ax[1].set_xlabel('Radius [au]', fontsize=MEDIUM_SIZE-2)
    ax[0].set_ylabel(clabel, **kw_ylabel)
    ax[1].set_ylabel('Residuals', **kw_ylabel) #+clabel_res.split('residuals')[-1] #--> unit

    make_1d_legend(ax[0], ncol=2)

    make_substructures(ax[0], gaps=gaps, rings=rings, label_gaps=True, label_rings=True)  
    make_substructures(ax[1], gaps=gaps, rings=rings)
    
    plt.savefig('radial_profile_%s.png'%tag_base, bbox_inches='tight', dpi=200)
    show_output(args)



from discminer.core import Data
from discminer.rail import Contours
from discminer.plottools import (make_up_ax,
                                 mod_major_ticks,
                                 mod_minor_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

from utils import (get_2d_plot_decorators,
                   get_noise_mask,
                   load_moments,
                   load_disc_grid,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import json
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot moment or residual maps in 3D', description='')
parser.add_argument('-t', '--type', default='residuals', type=str,
                    choices=['data', 'model', 'residuals'],
                    help="Visualise main observables in 3D. DEFAULTS to 'residuals'")
parser.add_argument('-d', '--density', default=1, type=int,
                    help="Total points shown: ntot/args.density. DEFAULTS to 1")
args = add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, smooth=True)
     
#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

if args.type=='residuals':
    clim0 = -clim
    clim1 = clim
else:    
    clim0 = levels_im[0]    
    clim1 = levels_im[-1]
    
#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout
au_to_m = u.au.to('m')

#********************
#LOAD DATA AND GRID
#********************
datacube = Data(file_data, dpc) # Read data and convert to Cube object
noise_mean, mask = get_noise_mask(datacube)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

#*************************
#LOAD MOMENT MAPS    
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

cmap = cmap_mom
if args.type=='residuals':
    map2d = residuals
    cmap = cmap_res
elif args.type=='data':
    
    map2d = moment_data
elif args.type=='model':
    map2d = moment_model

#**************************
#MAKE PLOT
fig = plt.figure(figsize=(14, 9))
ax = plt.axes(projection='3d')

xu = (R['upper']*np.cos(phi['upper'])).flatten()/au_to_m
yu = (R['upper']*np.sin(phi['upper'])).flatten()/au_to_m
zu = z['upper'].flatten()/au_to_m

xl = (R['lower']*np.cos(phi['lower'])).flatten()/au_to_m
yl = (R['lower']*np.sin(phi['lower'])).flatten()/au_to_m
zl = z['lower'].flatten()/au_to_m

di = args.density

ax.scatter3D(xu[::di], yu[::di], zu[::di], c=map2d.flatten()[::di], s=2.5, cmap=cmap, vmin=clim0, vmax=clim1)
ax.set_box_aspect([4,4,0.7])
ax.set_title(ctitle + r' $-$ ' + clabel + r' $-$ ' + args.type.capitalize(), pad=5, fontsize=15)
plt.show()
#plt.savefig('moment+residuals_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)





"""
Useful repeating code
"""
import numpy as np
import argparse
import runpy
import sys
import os

path_mining = os.path.dirname(os.path.realpath(__file__)) + '/mining/'

def _check_and_return_parser(parserobj, prog='', description=''):
    if parserobj is None:
        parser = argparse.ArgumentParser(prog=prog, description=description)
    elif isinstance(parserobj, argparse._SubParsersAction):
        parser = parserobj.add_parser(prog, help=description)
    elif isinstance(parserobj, ArgumentParser):
        parser = parserobj
    else:
        print("Please provide a valid parser object from the argparse library...")
        sys.exit()
    return parser

#Customised action (from https://stackoverflow.com/questions/25295487/python-argparse-value-range-help-message-appearance)
class Range(argparse.Action):
    def __init__(self, minimum=None, maximum=None, *args, **kwargs):
        self.min = minimum
        self.max = maximum
        kwargs["metavar"] = "[%d, %d]" % (self.min, self.max)
        super(Range, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_string=None):
        value = np.asarray(value)
        ind = (value >= self.min) & (value <= self.max)
        if not ind.all():
            msg = 'invalid choice: %r (choose value from [%d, %d])' % (value, self.min, self.max)
            raise argparse.ArgumentError(self, msg)
        setattr(namespace, self.dest, value)
        
#**************************************
#DEFINE AND CUSTOMISE MINING ARGUMENTS
#-------------------->
def _mining_channels(parserobj, prog='channels', description='Make model channel maps and compare to data'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-mb', '--make_beam', default=-1, choices=[-1,0,1], type=int,
                        help="Convolve by beam? Defaults to -1 (convolve if downsampling size used for fit < beam size)")
    #Define and parse additional arguments. 'True' means 'enable' the argument, they can default to different values; use -h for help.
    add_parser_args(parser, planck=True, Rinner=0.0, Router=1.2, absolute_Rinner=True, absolute_Router=True, writefits=1) 
    return parser

def _mining_moments1d(parserobj, prog='moments1d', description='Make (gaussian, bell, or quadratic) moment maps and save output into .fits files'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-pk', '--peakkernel', default=1, type=int,
                        help="Return peak from fitted kernel (1) or global peak intensity (0). DEFAULTS to 1 (Return peak from kernel).")
    parser.add_argument('-fc', '--fit_continuum', default=0, type=int,
                        help="Fit continuum and save into .fits file? Defaults to 0")
    parser.add_argument('-fdata', '--fit_data', default=1, type=int,
                        help="Fit datacube and save moments into .fits files? Defaults to 1")
    parser.add_argument('-fmodel', '--fit_model', default=1, type=int,
                        help="Fit modelcube and save moments into .fits files? Defaults to 1")    
    add_parser_args(parser, kernel=True, planck=True, sigma=3)
    return parser

def _mining_moments2d(parserobj, prog='moments2d', description='Make (double Gauss or double Bell) moment maps and save output into .fits files'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-ni', '--niter', default=10, type=int,
                        help="Number of iterations to re-do fit on hot pixels. DEFAULS to 10")
    parser.add_argument('-ne', '--neighs', default=5, type=int,
                        help="Number of neighbour pixels on each side of hot pixel considered for the iterative fit. DEFAULS to 5")
    parser.add_argument('-fdata', '--fit_data', default=1, type=int,
                        help="Fit datacube and save moments into .fits files? Defaults to 1")
    parser.add_argument('-fmodel', '--fit_model', default=1, type=int,
                        help="Fit modelcube and save moments into .fits files? Defaults to 1")
    add_parser_args(parser, kernel='doublebell', kind=True, planck=True, sigma=3)
    return parser

def _mining_moment_residuals(parserobj, prog='moment+residuals', description='Show Data vs Model moment map and residuals'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    add_parser_args(parser, moment=True, kernel=True, kind=True, sigma=3, surface=True, smooth=True, mask_phi=True, mask_R=True, Router=True, radius_planet=True, phi_planet=True, label_planet=True, input_coords=True)  
    return parser

def _mining_moment_offset(parserobj, prog='moment+offset', description='Show moment map and a zoom-in illustrating offset from the centre'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-scontours', '--show_contours', default=1, type=int, help="Overlay moment map contours? DEFAULTS to 1.")
    parser.add_argument('-zoom', '--zoom_size', default=100, type=float, help="Physical size of the zoom-in region. DEFAULTS to 100 au.")
    add_parser_args(parser, moment=True, kernel=True, kind=True, sigma=True, show_continuum='none', surface=True, smooth=True,
                    radius_planet=True, phi_planet=True, label_planet=True, input_coords=True)
    return parser

def _mining_residuals_deproj(parserobj, prog='residuals+deproj', description='Show residuals from a moment map, deprojected onto the disc reference frame'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    add_parser_args(
        parser,
        moment=True, kernel=True, kind=True, sigma=3, surface=True, projection=True, Rinner=True, Router=0.95, absolute_Rinner=True, absolute_Router=True, smooth=True,
        show_continuum=True, show_filaments=True, spiral_ids=True, spiral_type=True, spiral_moment=True, colorbar=True,
        mask_R=True, mask_phi=True, radius_planet=True, phi_planet=True, label_planet=True, input_coords=True
    )
    return parser

def _mining_residuals_all(parserobj, prog='residuals+all', description='Show ALL moment map residuals, deprojected onto the sky OR disc reference frame'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-c', '--coords', default='sky', type=str, choices=['disc', 'disk', 'sky'], help="Reference frame in which the maps will be displayed. DEFAULTS to 'sky'.")
    add_parser_args(parser, kernel=True, kind=True, sigma=3, surface=True, Rinner=True, Router=True, smooth=True, radius_planet=True, phi_planet=True, label_planet=True, input_coords=True)    
    return parser

def _mining_pick(parserobj, prog='pick', description='Use Pick tools. Fold residual maps and identify peak and clustered residuals'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-c', '--clean_thres', default=10, type=float, help="Sigma threshold above which peak residuals will be rejected. DEFAULTS to 10.")
    parser.add_argument('-sp', '--show_peaks', default=1, type=int, help="Show peak residuals on 2D maps. DEFAULTS to 1.")
    parser.add_argument('-sglobal', '--show_global', default=1, type=int, help="Show global peak. DEFAULTS to 1.")
    parser.add_argument('-pk', '--percentage_kepler', default=0, type=int, help="Refer velocity residuals to the local Keplerian rotation. DEFAULTS to 0.")
    parser.add_argument('-clusters', '--clusters', default=0, type=int, help="Make clusters. DEFAULTS to 0.")
    parser.add_argument('-np', '--nphi', default=6, type=int, help="Number of azimuthal clusters. DEFAULTS to 6.")
    parser.add_argument('-nr', '--nr', default=6, type=int, help="Number of radial clusters. DEFAULTS to 6.")
    parser.add_argument('-sleg', '--show_legend', default=0, type=int, help="Show markers legend. DEFAULTS to 0.")
    add_parser_args(
        parser,
        moment=True, kernel=True, kind=True, surface=True, fold=True, projection=True, Rinner=3.0, Router=0.9, absolute_Rinner=True, absolute_Router=True,
        colorbar=True, smooth=True, mask_phi=True, mask_R=True, radius_planet=True, phi_planet=True, label_planet=True, input_coords=True, writetxt=1, writefits=0
    )
    return parser

def _mining_radial_profiles(parserobj, prog='radprof', description='Extract and show radial profiles from moment maps AND residuals'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-sf', '--savgol_filter', default=0, type=int,
                        help="Apply a Savitzky-Golay filter to smooth out the plotted curves. DEFAULTS to False.")
    parser.add_argument('-vp', '--vphi_discminer', default=1, type=int,
                        help="Subtract model vphi or perfect Keplerian vphi? DEFAULTS to True (i.e. subtract model vphi).")
    parser.add_argument('-ig', '--interpgrid', default=0, type=int,
                        help="Use linear grid interpolation for radial/azimuthal profile extraction. DEFAULTS to False (i.e. use native grid).")
    add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, writetxt=True, mask_minor=True, mask_major=True, mask_R=True, mask_phi=True,
                    Rinner=True, absolute_Rinner=True, Router=True, absolute_Router=True, sigma=True, smooth=True)
    return parser

def _mining_radial_profiles_wedge(parserobj, prog='radprof+wedge', description='Extract and show radial profiles from moment residuals within specific wedges'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-wedges', '--wedges', nargs='*', default=[-60, 0, 0, 60, 120,  180, -180, -120], type=float, minimum=-180, maximum=180, action=Range,
                        help="Azimuthal boundaries of the wedges where radial profiles will be extracted. USAGE: -wedges 30 40 -60 -40 means two wedges, with boundaries (30, 40) and (-60, -40) deg. DEFAULTS to -60, 0, 0, 60, 120,  180, -180, -120.")
    parser.add_argument('-sf', '--savgol_filter', default=0, type=int,
                        help="Apply a Savitzky-Golay filter to smooth out the plotted curves. DEFAULTS to False.")
    add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, Rinner=True, absolute_Rinner=True, Router=True, absolute_Router=True, sigma=np.inf, smooth=True)
    return parser

def _mining_azimuthal_profiles(parserobj, prog='azimprof', description='Extract and show azimuthal profiles from moment maps OR residuals'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-r', '--radius', default=100, type=float,
                        help="Reference annulus shown as a solid black line. DEFAULTS to 100 au")
    parser.add_argument('-t', '--type', default='residuals', type=str, choices=['data', 'model', 'residuals'],
                        help="Compute profiles on data, model or residual moment map. DEFAULTS to 'residuals'")
    parser.add_argument('-ig', '--interpgrid', default=0, type=int,
                        help="Use linear grid interpolation for radial/azimuthal profile extraction. DEFAULTS to False (i.e. use native grid).")
    add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, Rinner=True, Router=True, smooth=True, writetxt=True)
    return parser

def _mining_spectra(parserobj, prog='spectra', description='Extract and show line profiles along a specific annulus, every 30 deg'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-r', '--radius', default=200, type=float, help="Annulus where line profiles will be extracted. DEFAULTS to 200 au")
    parser.add_argument('-f', '--showfit', default=1, type=int, choices=[0, 1], help="Overlay best-fit line profile? DEFAULTS to 1")
    parser.add_argument('-ch', '--channel_id', default=-1, type=int, help="Id of intensity channel map to show. DEFAULTS to -1 (if -1, show moment map passed through -m flag)")
    parser.add_argument('-phi0', '--phi0', default=0, type=float, help="Starting azimuthal location for line profile extraction. DEFAULTS to 0 deg")
    parser.add_argument('-np', '--npix', default=0, type=int, help="Number of pixels around central pixel considered for line profile extraction. DEFAULS to 0")
    parser.add_argument('-t', '--type', default='data', type=str, choices=['data', 'model'], help="Show line profiles from data or model? DEFAULTS to 'data'")
    parser.add_argument('-scontours', '--show_contours', default=1, type=int, help="Overlay contours. DEFAULTS to 1.")    
    add_parser_args(parser, moment='peakintensity', kernel=True, kind=True, smooth=True, planck=True)
    return parser

def _mining_channels_peakint(parserobj, prog='channels+peakint', description='Show Data vs Model channel maps, peak intensities, and residuals'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-nc', '--nchans', default=5, type=int, help="Number of channels to plot")
    parser.add_argument('-st', '--step', default=4, type=int, help="Plot every #step channels")
    add_parser_args(parser, sigma=3, moment='peakintensity', kernel=True, surface=True, smooth=True, Rinner=True, Router=1.1)
    return parser

def _mining_isovelocities(parserobj, prog='isovelocities', description='Show Data vs Model isovelocity contours'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    add_parser_args(
        parser,
        moment=True, kernel=True, kind=True, surface=True, projection=True, Rinner=True, Router=0.95, absolute_Rinner=True, absolute_Router=True, smooth=True,
        show_continuum=True, show_filaments=True, spiral_ids=True, spiral_type=True, spiral_moment=True,
        mask_R=True, mask_phi=True, radius_planet=True, phi_planet=True, label_planet=True, input_coords=True
    )
    return parser

def _mining_pv_diagram(parserobj, prog='pv', description='Show PV diagram extracted along a specific axis'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-pvphi', '--pvphi', default=0.0, type=float, help="Azimuth of the axis along which the PV diagram is to be extracted. DEFAULTS to 0 deg")
    add_parser_args(parser, moment='peakintensity', kernel=True, kind=True, surface=True, smooth=True, Rinner=True, Router=1)
    return parser

def _mining_attributes(parserobj, prog='attributes', description='Show model attributes (z, v, I, Lw) as a function of radius'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    add_parser_args(parser)
    return parser

def _mining_gradient(parserobj, prog='gradient', description='Show peak, radial AND/OR azimuthal gradient from residual maps'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-gt', '--threshold', default=2, type=float, help="Minimum gradient value a peak must have to be taken into account. Defaults to 2 (m/s/au).")
    parser.add_argument('-sleg', '--show_legend', default=0, type=int, help="Show markers legend. DEFAULTS to 0.")
    add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, Rinner=True, absolute_Rinner=True, Router=True, absolute_Router=True,
                    gradient=True, smooth=True, radius_planet=True, phi_planet=True, label_planet=True, input_coords='disc')
    return parser

def _mining_parcube(parserobj, prog='parcube', description='Show cube reconstructed from fit parameters vs data cube'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    add_parser_args(parser, kernel=True, kind=True, surface='both')
    return parser

def _mining_parfile(parserobj, prog='parfile', description='Make JSON parameter file based on input log_pars.txt and prepare_data.py'):
    parser = _check_and_return_parser(parserobj, prog=prog, description=description)
    parser.add_argument('-dir_log', '--dir_log', default='./', type=str, help="Directory where log_pars.txt is located. DEFAULTS to './'")
    parser.add_argument('-dir_data', '--dir_data', default='./', type=str, help="Directory where the input datacube.fits is located. DEFAULTS to './'")
    parser.add_argument('-dir_model', '--dir_model', default='./', type=str, help="Directory where model files will be stored. DEFAULTS to './'")    
    parser.add_argument('-f', '--log_file', default='', type=str, help="If not specified, try to guess input log_pars.txt with existing files. DEFAULTS to ''.")
    parser.add_argument('-p', '--prepare_data', default='prepare_data.py', type=str, help="Script employed to clip and downsample the cube of interest. DEFAULTS to prepare_data.py.")
    parser.add_argument('-j', '--json_file', default='parfile.json', type=str, help="Name of output JSON file. DEFAULTS to parfile.json.")
    parser.add_argument('-o', '--overwrite', default=0, type=int, help="overwrite if parfile.json exists. DEFAULTS to 0.")
    parser.add_argument('-r', '--reset', default=0, type=int,
                        help="If (1) AND --overwrite, rewrite parfile.json. If (0) AND --overwrite, forward 'custom' dictionary, and rewrite metadata and model parameters only. DEFAULTS to 0.")
    parser.add_argument('-d', '--download_cube', default=0, type=int, help="Download reduced ready-to-use cube from the NRAO server (valid for exoALMA data). DEFAULTS to 0.")
    add_parser_args(parser)


_mining_parser_func = {
    'parfile': _mining_parfile,    
    'channels': _mining_channels,
    'moments1d': _mining_moments1d,
    'moments2d': _mining_moments2d,
    'parcube': _mining_parcube,
    'channels+peakint': _mining_channels_peakint,    
    'attributes': _mining_attributes,
    'radprof': _mining_radial_profiles,
    'radprof+wedge': _mining_radial_profiles_wedge,    
    'azimprof': _mining_azimuthal_profiles,    
    'spectra': _mining_spectra,    
    'moment+residuals': _mining_moment_residuals,
    'moment+offset': _mining_moment_offset,
    'residuals+deproj': _mining_residuals_deproj,
    'residuals+all': _mining_residuals_all,
    'pick': _mining_pick,
    'gradient': _mining_gradient,
    'isovelocities': _mining_isovelocities,
    'pv': _mining_pv_diagram,
}

scripts = {
    'parfile': 'make_parfile.py',
    'channels': 'make_channels.py',
    'moments1d': 'make_single_moments.py',
    'moments2d': 'make_double_moments.py',
    'parcube': 'show_parcube.py',        
    'channels+peakint': 'plot_channels+peakint.py',    
    'attributes': 'plot_attributes_model.py',
    'radprof': 'plot_radial_profiles.py',
    'radprof+wedge': 'plot_radial_profiles+wedge.py',    
    'azimprof': 'plot_azimuthal_profiles.py',    
    'spectra': 'plot_spectra.py',    
    'moment+residuals': 'plot_moment+residuals.py',
    'moment+offset': 'plot_moment+offset.py',
    'residuals+deproj': 'plot_residuals+deproj.py',
    'residuals+all': 'plot_residuals+all.py',
    'pick': 'plot_peak_residuals.py',
    'gradient': 'plot_gradient.py',
    'isovelocities': 'plot_isovelocities.py',
    'pv': 'plot_pv_diagram.py',
}

#<----------------------

#*********************************
#PARSE ARGUMENTS AND CALL SCRIPTS
#*********************************
def main():
    parser = argparse.ArgumentParser(description="****>Run mining analysis scripts<****")
    subparsers = parser.add_subparsers(dest='make', help="Select a subscript to run.")

    for key in scripts:
        _mining_parser_func[key](subparsers) #Append subparsers

    args = parser.parse_args()
    _adjust_args(args)
    options = argparse.Namespace(**vars(args)) #Organised set of arguments passed as globals to the script
    
    if args.make in scripts:
        script_path = os.path.join(path_mining, scripts[args.make])
        script_args = {k: v for k, v in vars(args).items() if v is not None and k not in ('make',)}
        
        print(f"Running {script_path} with arguments {script_args}")

        #Run the selected script, passing the arguments as globals
        runpy.run_path(script_path, init_globals={'args': options} )
        
    else:
        print("Please specify a valid subcommand. Use -h for help.")

        
def add_parser_args(parser,
                    show_output=True, show_block=True,
                    moment=False, kernel=False, surface=False, smooth=False,
                    Rinner=False, Router=False, absolute_Rinner=False, absolute_Router=False,
                    kind=False, projection=False,
                    planck=False,
                    sigma=False,
                    mask_minor=False, mask_major=False, 
                    mask_phi=False, mask_R=False,
                    fold=False,
                    writetxt=False, writefits=True,
                    gradient=False,
                    colorbar=False,
                    show_continuum=False, show_filaments=False,
                    spiral_ids=False, spiral_type=False, spiral_moment=False,
                    filament_ids=False, filament_moment=False,
                    radius_planet=False, phi_planet=False, label_planet=False, input_coords=False,
):

    def set_default(val, default):
        if isinstance(val, bool):
            return default
        else:
            return val
        
    if moment is not False:
        d0 = set_default(moment, 'velocity')
        parser.add_argument('-m', '--moment', default=d0, type=str,
                            choices=['velocity', 'linewidth', 'lineslope', 'peakint', 'peakintensity', 'v0r', 'v0phi', 'v0z', 'vr_leftover', 'delta_velocity', 'delta_linewidth', 'delta_peakintensity', 'reducedchi2'],
                            help="Type of moment map to be analysed. DEFAULTS to '%s'"%d0)

    if kernel is not False:
        d0 = set_default(kernel, 'gaussian')
        parser.add_argument('-k', '--kernel', default=d0, type=str,
                            choices=['gauss', 'gaussian', 'bell', 'quadratic', 'dgauss', 'doublegaussian', 'dbell', 'doublebell'],
                            help="Kernel used for line profile fit and calculation of moment maps. DEFAULTS to '%s'"%d0)
        
    if kind is not False:
        d0 = set_default(kind, 'mask')
        parser.add_argument('-ki', '--kind', default=d0, type=str, choices=['mask', 'sum'],
                            help="Method for merging upper and lower surface kernel profiles. DEFAULTS to '%s'"%d0)
        
    if surface is not False:
        d0 = set_default(surface, 'upper')
        parser.add_argument('-s', '--surface', default=d0, type=str,
                            choices=['up', 'upper', 'low', 'lower', 'both'],
                            help="Use upper or lower surface moment map. DEFAULTS to '%s'"%d0)
    
    if show_continuum is not False:
        d0 = set_default(show_continuum, 'none')
        parser.add_argument('-sc', '--show_continuum', default=d0, type=str,
                            choices=['none', 'all', 'band7', 'scattered'],
                            help="Which continuum map to show. DEFAULTS to '%s'"%d0)

    if show_filaments is not False:
        d0 = set_default(show_filaments, 0)
        parser.add_argument('-sf', '--show_filaments', default=d0, type=int,
                            choices=[0, 1],
                            help="Make and show filaments? DEFAULTS to '%s'"%d0)        
    if show_output is not False:
        d0 = set_default(show_output, 1)
        parser.add_argument('-so', '--show_output', default=d0, type=int,
                            choices=[0, 1],
                            help="Show output plots? DEFAULTS to '%s'"%d0)        

    if show_block is not False:
        d0 = set_default(show_block, 1)
        parser.add_argument('-sb', '--show_block', default=d0, type=int,
                            choices=[0, 1],
                            help="Block ipython session when a figure is displayed? DEFAULTS to '%s'"%d0)        
        
    if mask_minor is not False:
        d0 = set_default(mask_minor, 30.0)
        parser.add_argument('-b', '--mask_minor', default=d0, type=float,
                            help="+- azimuthal mask around disc minor axis for calculation of vphi and vz velocity components. DEFAULTS to %.1f deg"%d0)

    if mask_major is not False:
        d0 = set_default(mask_major, 30.0)
        parser.add_argument('-a', '--mask_major', default=d0, type=float,
                            help="+- azimuthal mask around disc major axis for calculation of vR velocity component. DEFAULTS to %.1f deg"%d0)

    if mask_phi is not False:
        d0 = set_default(mask_phi, [])
        parser.add_argument('-mask_phi', '--mask_phi', nargs='*', default=d0, type=float, help="Azimuthal boundaries of mask(s) to apply to data and model products. FORMAT: -mask_phi 30 40 -60 -40 means two masks, with boundaries (30, 40) and (-60, -40) deg. DEFAULTS to [].")

    if mask_R is not False:
        d0 = set_default(mask_R, [])
        parser.add_argument('-mask_R', '--mask_R', nargs='*', default=d0, type=float, help="Radial boundaries of mask(s) to apply to data and model products. FORMAT: -mask_R 50 75 30 60 means two masks, with boundaries (50, 75) and (30, 60) au . DEFAULTS to [].")                

    if fold is not False:
        d0 = set_default(fold, 'absolute')        
        parser.add_argument('-f', '--fold', default=d0, type=str,
                            choices=['absolute', 'standard'],
                            help="if moment=velocity, fold absolute or standard velocity residuals. DEFAULTS to '%s'"%d0)

    if projection is not False:
        d0 = set_default(projection, 'cartesian')                
        parser.add_argument('-p', '--projection', default=d0, type=str,
                            choices=['cartesian', 'polar'],
                            help="Project residuals onto a cartesian or a polar map. DEFAULTS to '%s'"%d0)

    if planck is not False:
        d0 = set_default(planck, 0)                
        parser.add_argument('-planck', '--planck', default=d0, type=int,
                            choices=[0, 1],
                            help="Use full Planck's law to convert Jy/bm to K. Defaults to 0 (i.e. assume RJ)")
        
    if writetxt is not False:
        d0 = set_default(writetxt, 1)                
        parser.add_argument('-w', '--writetxt', default=d0, type=int,
                            choices=[0, 1],
                            help="Write output into txt file(s)? DEFAULTS to %d"%d0)

    if writefits is not False:
        d0 = set_default(writefits, 0)                
        parser.add_argument('-wf', '--writefits', default=d0, type=int,
                            choices=[0, 1],
                            help="Write output into fits file(s)? DEFAULTS to %d"%d0)
        
    if Rinner is not False:
        d0 = set_default(Rinner, 1.0)                        
        parser.add_argument('-i', '--Rinner', default=d0, type=float,
                            help="Number of beams to mask out from the disc inner region. DEFAULTS to %.2f"%d0)

    if absolute_Rinner is not False:
        d0 = set_default(absolute_Rinner, -1)
        parser.add_argument('-ai', '--absolute_Rinner', default=d0, type=float,
                            help="If >= 0, assume this absolute value instead of Rinner to set the inner radius of the analysis domain. Defaults to -1.")
        
    if Router is not False:
        d0 = set_default(Router, 0.98)                
        parser.add_argument('-o', '--Router', default=d0, type=float,
                            help="Fraction of Rout to consider as the disc outer radius for the analysis. DEFAULTS to %.2f"%d0)

    if absolute_Router is not False:
        d0 = set_default(absolute_Router, -1)
        parser.add_argument('-ao', '--absolute_Router', default=d0, type=float,
                            help="If >= 0, assume this absolute value instead of Router to set the outer radius of the analysis domain. Defaults to -1.")
                
    if sigma is not False:
        d0 = set_default(sigma, 5)                
        parser.add_argument('-si', '--sigma', default=d0, type=float,
                            help="Mask out pixels with values below sigma threshold. DEFAULTS to %.1f"%d0)

    if gradient is not False:
        d0 = set_default(gradient, 'r')
        parser.add_argument('-g', '--gradient', default=d0, type=str, choices=['peak', 'r', 'phi'],
                            help="Coordinate along which the gradient will be computed. If 'peak', the maximum gradient is computed. DEFAULTS to '%s'"%d0)

    if colorbar is not False:
        d0 = set_default(colorbar, 1)
        parser.add_argument('-cbar', '--colorbar', default=d0, type=int, help="Show colorbar. DEFAULTS to %d"%d0)        
        
    if smooth is not False:
        d0 = set_default(smooth, 0.0)
        parser.add_argument('-sm', '--smooth', default=d0, type=float,
                            help="Smooth up moment_data using scipy.ndimage gaussian_filter? DEFAULTS to %.1f (characteristic size of smoothing kernel in pixels)"%d0)        

    if spiral_ids is not False:
        d0 = set_default(spiral_ids, [])
        parser.add_argument('-spids', '--spiral_ids', nargs='*', default=d0, type=int, help="Which filament should be fitted with a spiral function. Spirals are overlaid on the map and their best-fit parameters are saved into a txt file. FORMAT: -spids 1 3 -2. DEFAULTS to [].")

    if spiral_type is not False:
        d0 = set_default(spiral_type, 'linear')        
        parser.add_argument('-sptype', '--spiral_type', default=d0, choices=['linear', 'log'], help="Type of spiral functional form to fit filaments. DEFAULTS to %s."%d0)

    if spiral_moment is not False:
        d0 = set_default(spiral_moment, 'velocity')        
        parser.add_argument('-spmom', '--spiral_moment', default=d0, choices=['velocity', 'linewidth', 'peakint', 'peakintensity'], help="Moment map from which spirals were extracted. DEFAULTS to %s."%d0)        

    if filament_ids is not False:
        d0 = set_default(filament_ids, [])
        parser.add_argument('-filids', '--filament_ids', nargs='*', default=d0, type=int, help="Filament(s) to be shown. FORMAT: -filids 1 3 -2. DEFAULTS to [].")

    if filament_moment is not False:
        d0 = set_default(filament_moment, 'velocity')        
        parser.add_argument('-filmom', '--filament_moment', default=d0, choices=['velocity', 'linewidth', 'peakint', 'peakintensity'], help="Moment map from which filament(s) were extracted. DEFAULTS to %s."%d0)        

    if radius_planet is not False:
        d0 = set_default(radius_planet, [])                
        parser.add_argument('-rp', '--rp', nargs='*', default=d0, type=float, help="Mark radial location of planet(s) in au. DEFAULTS to [].")

    if phi_planet is not False:
        d0 = set_default(phi_planet, [])                
        parser.add_argument('-phip', '--phip', nargs='*', default=d0, type=float, help="Mark azimuthal location of planet(s) in degrees. DEFAULTS to [].")
        
    if label_planet is not False:
        d0 = set_default(label_planet, [])                
        parser.add_argument('-labelp', '--labelp', nargs='*', default=d0, type=str, help="Add label(s) centred on the planet(s) marker. DEFAULTS to [].")

    if input_coords is not False:
        d0 = set_default(input_coords, 'disc')
        parser.add_argument('-input_coords', '--input_coords', default=d0, type=str,
                            choices=['sky', 'disc', 'disk'],
                            help="Reference frame of the input planet coordinates. If 'sky', -rp must be the projected distance of the planet in arcsecs and -phip the Position Angle of the planet measured from the North. If 'disc' or 'disk', -rp must be the orbital radius of the planet in au and -phip the azimuthal location measured from the disc x axis. DEFAULTS to '%s'"%d0)

        
def _adjust_args(args):
    try:
        if args.moment=='peakint':
            args.moment = 'peakintensity'
    except AttributeError:
        pass

    try:
        if args.spiral_moment=='peakint':
            args.spiral_moment = 'peakintensity'
    except AttributeError:
        pass
    
    try:
        if args.surface=='up':
            args.surface = 'upper'
    except AttributeError:
        args.surface = 'upper'

    try:
        if args.surface=='low':
            args.surface = 'lower'
    except AttributeError:
        args.surface = 'upper'

    try:
        if args.kernel=='gauss':
            args.kernel = 'gaussian'
    except AttributeError:
        args.kernel = 'gaussian'
    
    try:
        if args.kernel=='dgauss':
            args.kernel = 'doublegaussian'
    except AttributeError:
        args.kernel = 'gaussian'

    try:
        if args.kernel=='dbell':
            args.kernel = 'doublebell'
    except AttributeError:
        args.kernel = 'gaussian'
    

if __name__ == '__main__':
    main()

    



from discminer.plottools import (get_discminer_cmap,
                                 get_continuous_cmap,
                                 make_pie_map,
                                 make_substructures,
                                 use_discminer_style)

from utils import (init_data_and_model,
                   get_2d_plot_decorators,
                   get_noise_mask,
                   load_moments,
                   add_parser_args)

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from astropy import units as u

import json
import copy
from argparse import ArgumentParser

use_discminer_style()

parser = ArgumentParser(prog='plot residual maps', description='Plot residual maps')
args = add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, projection=True, smooth=True)
     
#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Rout = best['intensity']['Rout']

gaps = custom['gaps']
rings = custom['rings']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube)
vchannels = datacube.vchannels
model.make_model()

#****************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

#******************
#LOAD DISC GEOMETRY
ng = model.grid['nx']
R_s = model.grid['R'].reshape(ng, ng)*u.m.to('au')
phi_s = model.grid['phi'].reshape(ng, ng)

X = R_s*np.cos(phi_s)
Y = R_s*np.sin(phi_s)

#****************
#USEFUL FUNCTIONS
def readtxt(tag):
    if tag!='': tag = '_'+tag
    return np.loadtxt('radial_profile_%s%s.dat'%(args.moment, tag)).T

def make_savgol(prof):
    try:
        ysav = savgol_filter(prof, 5, 3)
        ysav_deriv = savgol_filter(prof, 5, 3, deriv=1)
    except np.linalg.LinAlgError:
        ysav = prof
        ysav_deriv = None
    return ysav, ysav_deriv 

#************************
#LOAD VELOCITY COMPONENTS
Rp, vp, dvp = readtxt('vphi')
Rr, vr, dvr = readtxt('vr')
Rz, vz, dvz = readtxt('vz')    

f_vp = interp1d(Rp, 1e3*vp)
f_vr = interp1d(Rr, 1e3*vr)
f_vz = interp1d(Rz, 1e3*vz)

try:
    vp_sav, dvp_sav = make_savgol(1e3*vp)
    f_dvp = interp1d(Rp, dvp_sav)
except np.linalg.LinAlgError:
    f_dvp = interp1d(Rp, np.zeros_like(Rp)) 

#************************
#DEFINE PIE MAP ARGUMENTS

#Maps to plot per quadrant
mask = (R_s>np.max(Rp)) | (R_s<np.min(Rp)) 
R_s[mask] = np.nan

quadrant_map2d = {
    1: f_vz(R_s),
    2: f_vr(R_s),
    3: f_vp(R_s),
    4: f_dvp(R_s)
}

#Contourf levels per quadrant
nlev = 128
vlim_u = 1e3*clim
dvlim_u = 0.25*vlim_u

levels_vel = np.linspace(-vlim_u, vlim_u, nlev)
levels_div = np.linspace(-dvlim_u, dvlim_u, nlev)    
    
quadrant_levels = {
    1: levels_vel,
    2: levels_vel,
    3: levels_vel,
    4: levels_div,
}

#cmap per quadrant
cmap_vel = get_discminer_cmap('velocity')
hex_list = ['#%s'%tmp for tmp in ["901b22","c2242d","fdffff","a3bf50","778b3b"]]
cmap_div = get_continuous_cmap(hex_list)

quadrant_cmap = {
    1: cmap_vel,
    2: cmap_vel,
    3: cmap_vel,
    4: cmap_div,
}

#colorbar label per quadrant
quadrant_clabel = {
    1: r'$\upsilon_{\rm z}$ [m/s]',
    2: r'$\upsilon_{\rm R}$ [m/s]',
    3: r'$\Delta$$\upsilon_{\phi}$ [m/s]',
    4: r'${\delta{\Delta \upsilon_{\phi}}}/{\delta{R}}$'
}

#Call function
fig, ax = make_pie_map(
    X*u.au, Y*u.au, np.max(Rp)*u.au,
    quadrant_map2d=quadrant_map2d,
    quadrant_levels=quadrant_levels,
    quadrant_cmap=quadrant_cmap,
    quadrant_clabel=quadrant_clabel,    
    gaps=gaps, rings=rings
)
make_substructures(ax, gaps=gaps, rings=rings, label_rings=True, twodim=True)

ax.set_title(ctitle, fontsize=16, color='k')
fig.savefig('pie_velocity_components_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)
plt.show()
    
#MAKE 2D PLOT OF RESIDUALS-VPHI-VZ to illustrate uncertainty on vR
#SHOW 2D VELOCITY UNCERTAINTIES



from .cube import Cube
from .tools.utils import InputError
from .disc2d import Model
from .cart import *

import numpy as np

from astropy.io import fits
from astropy import units as u

from radio_beam import Beam

import copy

func_defaults = {
    'velocity_func': keplerian_vertical,
    'z_upper_func': z_upper_exp_tapered,
    'z_lower_func': z_lower_exp_tapered,
    'intensity_func': intensity_powerlaw_rout,
    'linewidth_func': linewidth_powerlaw,
    'lineslope_func': lineslope_powerlaw,
    'line_profile': line_profile_bell,
    'line_uplow': line_uplow_mask
}
                    
class ReferenceModel(Cube):
    def __init__(
            self,
            disc = 'disc1',
            mol = '12co',
            kind = ['mask'], #model decorators
            Rmin = 1.0,
            Rmax = 700*u.au,
            dpc = 100*u.pc,
            npix = 128,
            subpixels = 0,
            vchannels = np.linspace(-5.0, 5.0, 101),
            write_extent = True,
            beam = Beam(
                major=0.15*u.arcsec,
                minor=0.15*u.arcsec,
                pa=0*u.deg
            ),
            init_params = {},
            init_funcs = {},
            init_header = {},
            filename = 'referencecube.fits',
            parfile = None
    ): 
        """
        Initialise ReferenceModel. Inherits `~discminer.disc2d.Cube` properties and methods.
        
        Parameters
        ----------
        Rmax : `~astropy.units.Quantity`
            Maximum 1D extent of the sky grid in physical units
        
        dpc : `~astropy.units.Quantity`
            Distance to the disc.
        """

        #Init default functional forms and parameters
        self.funcs = {
            'velocity': keplerian_vertical,
            'z_upper' : z_upper_exp_tapered,
            'z_lower' : z_lower_exp_tapered,
            'intensity' : intensity_powerlaw_rout,
            'linewidth' : linewidth_powerlaw,
            'lineslope' : lineslope_powerlaw,
            'line_profile' : line_profile_bell,
            'line_uplow' : line_uplow_mask,                        
        }
        
        self.params = {
            'velocity': {
                'Mstar': 1.0,
                'vel_sign': 1,
                'vsys': 0
            },
            'orientation': {
                'incl': 0.7,
                'PA': 0.0,
                'xc': 0.0,
                'yc': 0.0
            },
            'intensity': {
                'I0': 0.5,
                'p': -1.5, 
                'q': 2.0,
                'Rout': 500
            },
            'linewidth': {
                'L0': 0.3, 
                'p': -0.5, 
                'q': -0.3
            }, 
            'lineslope': {
                'Ls': 2.0, 
                'p': 0.3, 
                'q': 0.0
            },
            'height_upper': {
                'z0': 40.0,
                'p': 1.0,
                'Rb': 500,
                'q': 2.0
            },
            'height_lower': {
                'z0': 20.0,
                'p': 1.0,
                'Rb': 500,
                'q': 2.0
            }
        }

        #Update dicts based on input kwargs
        for key in init_params:
            self.params[key].update(init_params[key])

        self.funcs.update(init_funcs)
            
        #Velocity channels
        nchan = len(vchannels)
        dchan = abs(vchannels[1]-vchannels[0])

        #Init beam for header
        if isinstance(beam, Beam):
            bmaj = beam.major
            bmin = beam.minor
            bpa = beam.pa
        elif beam is None:
            bmaj = 0.0*u.deg
            bmin = 0.0*u.deg
            bpa = 0.0*u.deg
        else:
            raise InputError(beam, "beam must be either None or radio_beam.Beam object")
        
        #Header
        dpix = np.arctan((Rmax.to(u.au) / (0.5*(npix-1))) / dpc.to(u.au)).to(u.deg)

        header = dict(
            SIMPLE  = True, 
            BITPIX  = -32, 
            NAXIS   = 3,                                                  
            NAXIS1  = npix,
            NAXIS2  = npix,
            NAXIS3  = nchan,
            EXTEND  = True,                                                  
            BMAJ    = bmaj.to(u.deg).value,
            BMIN    = bmin.to(u.deg).value,
            BPA     = bpa.to(u.deg).value,
            BTYPE   = 'Intensity',
            OBJECT  = 'ReferenceDiscminer',
            BUNIT   = 'Jy/beam',
            CTYPE1  = 'RA---SIN',        
            CRVAL1  = 245.0,
            CDELT1  = -dpix.value,
            CRPIX1  = int(0.5*npix)+1,
            CUNIT1  = 'deg',                                                            
            CTYPE2  = 'DEC--SIN',
            CRVAL2  = -30.0,
            CDELT2  = dpix.value,
            CRPIX2  = int(0.5*npix)+1,
            CUNIT2  = 'deg',
            CTYPE3  = 'VRAD', #/ Radio velocity (linear) 
            CRVAL3  = vchannels[0],
            CDELT3  = dchan,
            CRPIX3  = 1,
            CUNIT3  = 'km/s',                                                            
            RESTFRQ = 3.457959899000E+11, #/Rest Frequency (Hz) - Band 7
            SPECSYS = 'LSRK', #/Spectral reference frame 
            VELREF  = 257 #/1 LSR, 2 HEL, 3 OBS, +256 Radio
        )
        
        header.update(init_header)
        hdu = fits.PrimaryHDU()
        hdu.header.update(header)        
        data = np.zeros((nchan, npix, npix))
        
        Cube.__init__(self, data, hdu.header, vchannels, dpc, beam=beam, filename=filename, disc=disc, mol=mol, kind=kind, parfile=parfile)
        
        #Init model
        model = Model(self, Rmax=Rmax, Rmin=Rmin, write_extent=write_extent, prototype=True, subpixels=subpixels) 
        
        model.velocity_func = self.funcs['velocity']
        model.z_upper_func = self.funcs['z_upper']
        model.z_lower_func = self.funcs['z_lower']
        model.intensity_func = self.funcs['intensity']
        model.linewidth_func = self.funcs['linewidth']
        model.lineslope_func = self.funcs['lineslope']
        model.line_profile = self.funcs['line_profile']
        model.line_uplow = self.funcs['line_uplow']

        model.params = copy.copy(self.params)
        
        #UPDATE DATACUBE VALUES
        self.data = model.make_model(make_convolve=True, return_data_only=True)        
        self.model = model

        #UPDATE JSON INFO
        self.json_metadata = {            
            'Rmin': self.model.Rmin,
            'Rmax': self.model.Rmax
        }

        self.json_params = self.params
        self.json_funcs = self.funcs

        if filename is not None:
            self.writefits()

        if parfile is not None:
            self.make_parfile()

            
